/*
 *  Copyright (c) 2017 Maksim Shegay.
 *  The faculty of Computational Mathematics and Cybernetics of Lomonosov Moscow State University
 *  All rights reserved.
 *
 *  parMatt is licensed under the GNU public license version 2.0.
 */

#include <assert.h>
#include <math.h>
#include <mpi.h>

#include "MultipleAlignment.h"

#include "mpi_helpers.h"
#include "mpi_pairscores.h"
#include "mpi_pool.h"
#include "mpi_serializer.h"

static MultipleAlignment **mpi_init_als(PDBChain **chains, const int n)
{
	MultipleAlignment **alignments = (MultipleAlignment **)malloc(n * sizeof(MultipleAlignment *));
	assert(alignments);
	for (int i = 0; i < n; i++) {
		alignments[i] = CreateSingleStrandAlignment(Distill(chains[i], i));
	}
	return alignments;
}

static int *updater(int len, const int nprocs)
{
	int *buf = calloc(nprocs, sizeof(int));
	assert(buf);
	MPI_Allgather(&len, 1, MPI_INT, buf, 1, MPI_INT, MPI_COMM_WORLD);
	return buf;
}

static void mpi_pairwise(MultipleAlignment **als, snapshot_t *snap, pairscores_t **scores, processed_t *aligned, const int nthreads, const int rank)
{
	taskpool_ctx_t ctx = {
	    .processed = aligned,
	    .snap = snap,
	    .rank = rank,
	};
	pthread_mutex_t *lock = malloc(sizeof(pthread_mutex_t));
	assert(lock);
	pthread_mutex_init(lock, NULL);

	void *align(mpi_task_t t)
	{
		double t1 = MPI_Wtime();
		MultipleAlignment *al1 = als[t.al1];
		MultipleAlignment *al2 = als[t.al2];
		MultipleAlignment *ret = AlignAlignments(al1, al2, 0, 0, 0);
		pthread_mutex_lock(lock);
		pairscores_add(scores, t.al1, t.al2, ret->score);
		pthread_mutex_unlock(lock);
		double t2 = MPI_Wtime();
		mpi_log(2, 0, "pairwise, aligned %d %d in %1.5lf seconds", t.al1, t.al2, t2 - t1);
		return ret;
	}

	taskpool_run(nthreads, ctx, updater, align);

	pthread_mutex_destroy(lock);
	free(lock);
}

static void mpi_iterative(MultipleAlignment **als, processed_t *aligned, snapshot_t *snap, MultipleAlignment *ma, pairscores_t **scores, const int iteration, const int nthreads, const int rank)
{
	taskpool_ctx_t ctx = {
	    .processed = aligned,
	    .snap = snap,
	    .rank = rank,
	};

	void *align(mpi_task_t t)
	{
		double t1 = MPI_Wtime();
		MultipleAlignment *al = als[t.al1];
		double bestScore = 0;
		int best1 = 0, best2 = 0;
		for (int j = 0; j < al->numChains; j++) {
			for (int k = 0; k < ma->numChains; k++) {
				pairscores_t *s = pairscores_find(scores, ma->chains[k]->id, al->chains[j]->id);
				if (s->score > bestScore) {
					bestScore = s->score;
					best1 = j;
					best2 = k;
				}
			}
		}
		MultipleAlignment *ret = AlignAlignments(al, ma, best1, best2, 0);
		double t2 = MPI_Wtime();
		mpi_log(2, 0, "iteration %d, aligned %d in %1.5lf seconds", iteration, t.al1, t2 - t1);
		return ret;
	}
	taskpool_run(nthreads, ctx, updater, align);
}

static MultipleAlignment *mpi_bcast_best_ma(mpi_serializer_pool_t *pool, processed_t *p, int rank)
{
	struct best_t {
		double score;
		int rank;
	};

	struct best_t best = {
	    .score = -INFINITY,
	    .rank = rank,
	};
	int index = 0;
	for (int i = 0; i < p->len; i++) {
		MultipleAlignment *al = p->data[i];
		if (al->score > best.score || i == 0) {
			best.score = al->score;
			index = i;
		}
	}
	if (p->len > 0) {
		mpi_serializer_pool_reset(pool);
		mpi_ma_serialize(pool, p->data[index]);
	}
	struct best_t out;
	MPI_Allreduce(&best, &out, 1, MPI_DOUBLE_INT, MPI_MAXLOC, MPI_COMM_WORLD);
	size_t size = 0;
	MultipleAlignment *ret = NULL;
	if (rank == out.rank) {
		size = pool->offset;
		ret = p->data[index];
		p->data[index] = p->data[p->len - 1];
		p->len--;
	}
	MPI_Bcast(&size, 1, MPI_INT, out.rank, MPI_COMM_WORLD);
	if (rank != out.rank) {
		mpi_serializer_pool_ensure(pool, size);
	}
	MPI_Bcast(pool->ptr, size, MPI_CHAR, out.rank, MPI_COMM_WORLD);
	if (rank != out.rank) {
		ret = mpi_ma_deserialize(pool->ptr);
	}

	return ret;
}

static int cleanup(MultipleAlignment **als, int n, const int index1, const int index2)
{
	int deleted = 0;
	for (int i = n - 1; i >= 0; i--) {
		for (int j = 0; j < als[i]->numChains; j++) {
			if (als[i]->chains[j]->id != index1 && als[i]->chains[j]->id != index2) {
				continue;
			}

			CleanupAlignment(als[i]);
			n--;
			als[i] = als[n];
			deleted++;
			break;
		}
	}

	return deleted;
}

static void mpi_cleanup_best(MultipleAlignment **als, int *n_als, processed_t *p, MultipleAlignment *ma)
{
	int index1 = ma->chains[0]->id;
	int index2 = ma->chains[ma->numChains - 1]->id;

	*n_als -= cleanup(als, *n_als, index1, index2);

	p->len -= cleanup((MultipleAlignment **)p->data, p->len, index1, index2);
}

static void mpi_barrier_wrapper(const int iteration)
{
	double t1 = MPI_Wtime();
	MPI_Barrier(MPI_COMM_WORLD);
	double t2 = MPI_Wtime();
	if (iteration == -1) {
		mpi_log(2, 0, "iteration pairwise, waiting on barrier %1.5lf seconds", t2 - t1);
		return;
	}
	mpi_log(2, 0, "iteration %d, waiting on barrier %1.5lf seconds", iteration, t2 - t1);
}

MultipleAlignment *mpi_align(PDBChain **chains, int n_als, int nthreads)
{
	double start_time, end_time;
	start_time = MPI_Wtime();

	int rank, nprocs;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

	mpi_serializer_pool_t *pool = mpi_serializer_pool_create();

	MultipleAlignment **alignments = mpi_init_als(chains, n_als);
	processed_t *aligned = processed_init();
	pairscores_t *scores = NULL;

	double t1, t2;
	t1 = MPI_Wtime();
	snapshot_t *snap = snapshot_init(nprocs);

	snapshot_fill_pairwise(snap, n_als);

	mpi_pairwise(alignments, snap, &scores, aligned, nthreads, rank);
	mpi_pairscores_gather(&scores, rank, nprocs); /* sync */

	t2 = MPI_Wtime();
	mpi_log(0, 1, "pairwise time: %1.5lf seconds", t2 - t1);


	for (int i = 0; n_als > 1; i++) {
		t1 = MPI_Wtime();
		MultipleAlignment *ma = mpi_bcast_best_ma(pool, aligned, rank); /* sync */
		mpi_cleanup_best(alignments, &n_als, aligned, ma);
		snapshot_fill_iterative(snap, n_als);
		mpi_iterative(alignments, aligned, snap, ma, &scores, i, nthreads, rank);

		mpi_barrier_wrapper(i);
		alignments[n_als++] = ma;
		t2 = MPI_Wtime();
		mpi_log(0, 1, "iteration %d time: %1.5lf seconds", i, t2 - t1);
	}

	MultipleAlignment *ret = alignments[0];
	pairscores_free(&scores);
	processed_destroy(aligned);
	free(alignments);
	mpi_serializer_pool_destroy(pool);
	snapshot_destroy(snap);

	end_time = MPI_Wtime();
	mpi_log(0, 1, "alignment done in %1.5lf seconds", end_time - start_time);
	if (rank == MASTER) {
		return ret;
	}
	CleanupAlignment(ret);
	return NULL;
}
